@@ -1,5 +1,4 @@
 use std::{
-    cell::RefCell,
     collections::BTreeMap,
     iter::{self, zip},
     ops::Deref,
@@ -11,6 +10,7 @@ use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
     cached_method,
+    caching::FastUnboundedCache,
     eq_by_big_hash::EqByBigHash,
     impl_eq_by_big_hash, pycall, python_error_exception,
     rearrange_spec::{ExpandToSpecOrShape, OpShape, OpSize, RearrangeSpec},
@@ -133,6 +133,18 @@ pub fn expand_node(
     {
         return Ok(circuit); // in all identical case we can return same circuit for efficiency
     }
+    if inputs
+        .iter()
+        .zip(circuit.children())
+        .all(|(new, old)| new.shape() == old.shape())
+    {
+        // all shapes equivalent, so no need to expand or do anything fancy
+        // Just special casing for efficiency
+        // We can't use unwrap because we still want to propagate errors from
+        // stuff like dtype mismatches or constraint violations.
+        return circuit.map_children_idxs(|i| Ok(inputs[i].clone()));
+    }
+
     let batch_ranks: Vec<usize> = zip(circuit.children(), inputs)
         .filter_map(|(old, new)| new.info().rank().checked_sub(old.info().rank()))
         .collect();
@@ -847,12 +859,13 @@ impl<F: FnMut(CircuitRc) -> Option<CircuitRc>> ReplaceForExpand for ReplaceForEx
     }
 }
 
+#[derive(Debug, Clone)]
 pub struct ReplaceExpander<F: ReplaceForExpand> {
     replacer: F,
     // TODO: this holds onto array tensors forever!
     // We could try to use 'weak' for the cache, but this is a bit sad... (drops a bunch of stuff instantly)
     // Fix me!
-    cache: cached::UnboundCache<(HashBytes, HashBytes), CircuitRc>,
+    cache: FastUnboundedCache<(HashBytes, HashBytes), CircuitRc>,
 }
 
 impl<F: FnMut(CircuitRc) -> Option<CircuitRc>> ReplaceExpander<ReplaceForExpandFn<F>> {
@@ -861,6 +874,7 @@ impl<F: FnMut(CircuitRc) -> Option<CircuitRc>> ReplaceExpander<ReplaceForExpandF
     }
 }
 
+#[derive(Debug, Clone, Copy)]
 pub struct ReplaceForExpandNoop;
 
 impl ReplaceForExpand for ReplaceForExpandNoop {
@@ -872,12 +886,12 @@ impl ReplaceForExpand for ReplaceForExpandNoop {
     }
 }
 
-type MapReplaceExpander = ReplaceExpander<ReplaceForExpandNoop>;
+pub type MapReplaceExpander = ReplaceExpander<ReplaceForExpandNoop>;
 impl MapReplaceExpander {
     pub fn new_noop() -> Self {
         Self {
             replacer: ReplaceForExpandNoop,
-            cache: cached::UnboundCache::new(),
+            cache: FastUnboundedCache::default(),
         }
     }
 }
@@ -886,7 +900,7 @@ impl<F: ReplaceForExpand> ReplaceExpander<F> {
     pub fn new(replacer: F) -> Self {
         Self {
             replacer,
-            cache: cached::UnboundCache::new(),
+            cache: FastUnboundedCache::default(),
         }
     }
 
@@ -934,15 +948,6 @@ where
     ReplaceExpander::new_fn(replacer).replace_expand(circuit)
 }
 
-thread_local! {
-    static MAP_REPLACE_EXPANDER: RefCell<MapReplaceExpander> =
-        RefCell::new(MapReplaceExpander::new_noop());
-}
-
-pub fn map_replace_expand(circuit: CircuitRc, map: &ReplaceMapRc) -> Result<CircuitRc> {
-    MAP_REPLACE_EXPANDER.with(|f| f.borrow_mut().replace_expand_with_map(circuit, map))
-}
-
 #[test]
 fn check_map_override() {
     let m: BTreeMap<_, _> = [(1, 2), (3, 4), (5, 7), (3, 9)].into_iter().collect();
