@@ -48,6 +48,7 @@ use rr_util::{
     util::{
         arc_ref_clone, arc_unwrap_or_clone, AsOp, EinsumAxes, HashBytes, HashBytesToPy, NamedAxes,
     },
+    IndexSet,
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 pub use set_symbolic_shape::SetSymbolicShape;
@@ -55,7 +56,7 @@ use thiserror::Error;
 pub use variable_nodes::{DiscreteVar, StoredCumulantVar, Tag};
 
 use crate::{
-    module::substitute_all_modules,
+    module::{get_free_symbols, substitute_all_modules},
     print::{print_circuit_stats, PrintOptions},
 };
 pub mod cached_circuit_properties;
@@ -160,6 +161,8 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         *uuid::Uuid::nil().as_bytes()
     }
 
+    fn get_autoname(&self) -> Result<Option<String>>;
+
     // ==== default only section ===
     // FUNCTIONS BELOW HERE *shouldn't* be overridden by implementors!
     // (if you do implement, this won't be picked up on by CircuitNodeDefer (union types etc)!)
@@ -1144,6 +1147,13 @@ macro_rules! circuit_node_extra_impl {
 #[macro_export]
 macro_rules! circuit_node_auto_impl {
     ($the_uuid:literal) => {
+        $crate::circuit_node_auto_impl!($the_uuid, none);
+
+        fn get_autoname(&self) -> Result<Option<String>> {
+            Ok(self.auto_name())
+        }
+    };
+    ($the_uuid:literal, none) => {
         #[inline]
         fn info(&self) -> &$crate::CachedCircuitInfo {
             &self.info
@@ -1168,6 +1178,14 @@ macro_rules! circuit_node_auto_impl {
             $crate::CircuitRc(std::sync::Arc::new(self.c()))
         }
     };
+
+    ($the_uuid:literal, no_autoname) => {
+        $crate::circuit_node_auto_impl!($the_uuid, none);
+
+        fn get_autoname(&self) -> anyhow::Result<Option<String>> {
+            anyhow::bail!("this node type doesn't support autoname");
+        }
+    };
 }
 
 impl<T: CircuitNodeDefer> CircuitNodeSelfOnlyHash for T {
@@ -1231,6 +1249,10 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
     fn rc(self) -> CircuitRc {
         self.custom_rc()
     }
+
+    fn get_autoname(&self) -> Result<Option<String>> {
+        self.as_trait_obj().get_autoname()
+    }
 }
 
 #[derive(Clone, Default)]
@@ -1241,16 +1263,16 @@ pub struct CachedCircuitInfo {
     pub device_dtype: TorchDeviceDtypeOp,
     pub named_axes: NamedAxes,
     pub symbolic_size_constraints: SymbolicSizeConstraints,
+    pub free_symbols: IndexSet<Symbol>,
 }
 
 #[derive(Clone, Copy, Hash, Debug, Eq, PartialEq)]
 pub struct CircuitFlags(pub u8);
 impl CircuitFlags {
-    pub const IS_CONSTANT: CircuitFlags = CircuitFlags(0b00001);
-    pub const IS_EXPLICITLY_COMPUTABLE: CircuitFlags = CircuitFlags(0b00010);
-    pub const CAN_BE_SAMPLED: CircuitFlags = CircuitFlags(0b00100);
-    pub const NO_UNBOUND_SYMBOLS: CircuitFlags = CircuitFlags(0b01000);
-    pub const USE_AUTONAME: CircuitFlags = CircuitFlags(0b10000);
+    pub const IS_CONSTANT: CircuitFlags = CircuitFlags(0b0001);
+    pub const IS_EXPLICITLY_COMPUTABLE: CircuitFlags = CircuitFlags(0b0010);
+    pub const CAN_BE_SAMPLED: CircuitFlags = CircuitFlags(0b0100);
+    pub const USE_AUTONAME: CircuitFlags = CircuitFlags(0b1000);
 
     pub const NONE: CircuitFlags = CircuitFlags(0b0);
     pub fn check(self, other: CircuitFlags) -> bool {
@@ -1261,7 +1283,6 @@ impl CircuitFlags {
         CircuitFlags::IS_EXPLICITLY_COMPUTABLE
             | CircuitFlags::IS_CONSTANT
             | CircuitFlags::CAN_BE_SAMPLED
-            | CircuitFlags::NO_UNBOUND_SYMBOLS
             | CircuitFlags::USE_AUTONAME
     }
 }
@@ -1358,9 +1379,6 @@ impl CachedCircuitInfo {
     pub fn is_explicitly_computable(&self) -> bool {
         self.flags.check(CircuitFlags::IS_EXPLICITLY_COMPUTABLE)
     }
-    pub fn no_unbound_symbols(&self) -> bool {
-        self.flags.check(CircuitFlags::NO_UNBOUND_SYMBOLS)
-    }
     pub fn use_autoname(&self) -> bool {
         self.flags.check(CircuitFlags::USE_AUTONAME)
     }
@@ -2080,7 +2098,7 @@ impl PyCircuitBase {
 
     #[getter]
     fn no_unbound_symbols(&self) -> bool {
-        self.info().no_unbound_symbols()
+        get_free_symbols(&**self.0).is_empty()
     }
 
     #[getter]
@@ -2098,6 +2116,10 @@ impl PyCircuitBase {
         self.0.name()
     }
 
+    fn get_autoname(&self) -> Result<Option<String>> {
+        self.0.get_autoname()
+    }
+
     #[getter]
     fn symbolic_size_constraints(&self) -> SymbolicSizeConstraints {
         self.info().symbolic_size_constraints.clone()
@@ -2172,6 +2194,8 @@ impl PyCircuitBase {
     fn rank(&self) -> usize {
         self.0.info().rank()
     }
+
+    #[getter]
     fn ndim(&self) -> usize {
         self.rank()
     }
@@ -2216,6 +2240,10 @@ impl PyCircuitBase {
     fn dtype(&self) -> Option<TorchDtype> {
         self.0.info().device_dtype.dtype
     }
+    #[getter]
+    fn torch_dtype(&self) -> Option<PyObject> {
+        self.0.info().device_dtype.get_torch_dtype()
+    }
 
     fn get_compatible_device_dtype(&self) -> TorchDeviceDtype {
         get_compatible_dtype(&***self)
