@@ -1,7 +1,7 @@
 use std::{
     hash::{Hash, Hasher},
     iter::{once, zip},
-    // sync::{Arc, Weak},
+    ops::Deref,
 };
 
 use anyhow::{bail, Context, Result};
@@ -9,12 +9,14 @@ use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*, pyclass::CompareOp};
 use rr_util::{
     cached_method,
+    caching::FastUnboundedCache,
     py_types::use_rust_comp,
     pycall, python_error_exception,
     rearrange_spec::RearrangeSpec,
     symbolic_size::SymbolicSizeProduct,
     tensor_util::{right_align_shapes, Shape, TorchDeviceDtype},
     util::{counts_g_1, is_unique},
+    IndexSet,
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet, FxHasher};
 use thiserror::Error;
@@ -233,16 +235,66 @@ impl ModuleSpec {
         Ok((result, out_bytes))
     }
 
+    /// we could check this on module spec construct if we wanted to
+    pub fn check_no_ident_matching_issues(&self) -> Result<()> {
+        let my_idents: HashMap<_, _> = self
+            .arg_specs
+            .iter()
+            .map(|arg_spec| (arg_spec.symbol.ident(), &arg_spec.symbol))
+            .collect();
+        let sym_single_map = self
+            .circuit
+            .as_symbol()
+            .map(|x| [x.clone()].into_iter().collect());
+        let syms = sym_single_map
+            .as_ref()
+            .unwrap_or(&self.circuit.info().free_symbols);
+
+        for sym_here in syms {
+            if let Some(&my_sym) = my_idents.get(&sym_here.ident()) {
+                if my_sym != sym_here {
+                    bail!(
+                        SubstitutionError::FoundNEQFreeSymbolWithSameIdentification {
+                            sym: my_sym.clone(),
+                            matching_sym: sym_here.clone()
+                        }
+                    )
+                }
+            }
+        }
+
+        Ok(())
+    }
+
+    pub fn check_no_free_bound_by_nested(&self, nodes: &[CircuitRc]) -> Result<()> {
+        let key: (_, Vec<_>) = (
+            self.get_hashable(),
+            nodes.iter().map(|x| x.info().hash).collect(),
+        );
+        if NO_FREE_BOUND_BY_NESTED.with(|cache| cache.borrow().contains(&key)) {
+            return Ok(());
+        }
+        recur_check_no_free_bound_by_nested(
+            self.circuit.clone(),
+            &self.arg_specs.iter().map(|x| &x.symbol).collect::<Vec<_>>(),
+            &nodes.iter().collect::<Vec<_>>(),
+            &mut Default::default(),
+        )?;
+        NO_FREE_BOUND_BY_NESTED.with(|cache| cache.borrow_mut().insert(key));
+        Ok(())
+    }
+
     pub fn substitute(
         &self,
         nodes: &Vec<CircuitRc>,
         name_prefix: Option<String>,
         name_suffix: Option<String>,
     ) -> Result<CircuitRc> {
-        let key: (_, Vec<HashBytes>, Option<String>) = (
+        let key: (_, Vec<HashBytes>, Option<String>, _) = (
             self.get_hashable(),
             nodes.iter().map(|x| x.info().hash).collect(),
             name_prefix.clone(),
+            name_suffix.clone(),
         );
 
         if let Some(result) = MODULE_EXPANSIONS.with(|cache| {
@@ -261,9 +313,8 @@ impl ModuleSpec {
             return Ok(result);
         }
 
-        for (arg_spec, sub_for_sym) in self.arg_specs.iter().zip(nodes) {
-            check_valid_sub(&self.circuit, &arg_spec.symbol, sub_for_sym)?
-        }
+        self.check_no_ident_matching_issues()?;
+        self.check_no_free_bound_by_nested(nodes)?;
 
         let shapes = nodes.iter().map(|x| x.info().shape.clone()).collect();
         let (mut expanded_shape, rep_hashes) = self.expand_shape(&shapes)?;
@@ -343,16 +394,35 @@ impl ModuleSpec {
     }
 }
 
-#[pyfunction]
-pub fn all_free(circuit: CircuitRc) -> Vec<Symbol> {
-    // returns syms in circuit visit order (order maybe matters here)
-    let mut all_syms = Vec::new();
-    visit_circuit_unwrap(substitute_all_modules(circuit), |x| {
-        if let Some(sym) = x.as_symbol() {
-            all_syms.push(sym.clone())
+pub enum FreeSymbolsWrapper<'a> {
+    Sym(IndexSet<Symbol>),
+    Other(&'a IndexSet<Symbol>),
+}
+
+impl<'a> Deref for FreeSymbolsWrapper<'a> {
+    type Target = IndexSet<Symbol>;
+    fn deref(&self) -> &Self::Target {
+        match self {
+            Self::Sym(x) => x,
+            Self::Other(x) => x,
         }
-    });
-    all_syms
+    }
+}
+
+pub fn get_free_symbols(circuit: &Circuit) -> FreeSymbolsWrapper<'_> {
+    // returns syms in circuit visit order (order maybe matters here).
+    // The fact the ordering matters here is why we use indexmap in free_symbols
+    if let Some(sym) = circuit.as_symbol() {
+        FreeSymbolsWrapper::Sym([sym.clone()].into_iter().collect())
+    } else {
+        FreeSymbolsWrapper::Other(&circuit.info().free_symbols)
+    }
+}
+
+#[pyfunction]
+#[pyo3(name = "get_free_symbols")]
+pub fn py_get_free_symbols(circuit: CircuitRc) -> Vec<Symbol> {
+    get_free_symbols(&circuit).iter().cloned().collect()
 }
 
 #[pymethods]
@@ -423,10 +493,10 @@ impl ModuleSpec {
 
     #[staticmethod]
     #[args(check_unique_arg_names = "true")]
-    pub fn new_all_free(circuit: CircuitRc, check_unique_arg_names: bool) -> Result<Self> {
-        let arg_specs = all_free(circuit.clone())
-            .into_iter()
-            .map(|x| ModuleArgSpec::new(x, true, true, true))
+    pub fn new_free_symbols(circuit: CircuitRc, check_unique_arg_names: bool) -> Result<Self> {
+        let arg_specs = get_free_symbols(&circuit)
+            .iter()
+            .map(|x| ModuleArgSpec::new(x.clone(), true, true, true))
             .collect();
         let out =
             Self::new(circuit, arg_specs, true, false).expect("method should guarantee valid");
@@ -648,7 +718,7 @@ pub struct Module {
     pub nodes: Vec<CircuitRc>,
     #[pyo3(get)]
     pub spec: ModuleSpec,
-    cached_substituted: CircuitRc,
+    substitute_cached_info: Box<CachedCircuitInfo>, // box for size
     info: CachedCircuitInfo,
     #[pyo3(get)]
     name: Option<String>,
@@ -657,15 +727,43 @@ pub struct Module {
 impl Module {
     #[apply(new_rc_unwrap)]
     pub fn try_new(nodes: Vec<CircuitRc>, spec: ModuleSpec, name: Option<String>) -> Result<Self> {
-        let cached_substituted = spec
-            .substitute(&nodes, None, None)
-            .with_context(|| format!("module construction failed on substitute name={:?}", name))?;
+        // we need to check this here because we substitute modules below where other checks run
+        spec.check_no_free_bound_by_nested(&nodes).with_context(|| {
+            const MSG: &str = "(free sym bound by nested)";
+            format!("module construction failed because substitution isn't valid {MSG} (name={name:?})")
+        })?;
+        let mut substitute_cached_info = ModuleSpec {
+            // substituing all
+            circuit: substitute_all_modules(spec.circuit.clone()),
+            arg_specs: spec.arg_specs.clone(),
+        }
+        .substitute(&nodes, None, None)
+        .with_context(|| format!("module construction failed on substitute name={name:?}"))?
+        .info()
+        .clone();
+
+        // the substituted free symbols is *nearly* right, but has a few issues:
+        // - nodes might be unused which would result in missing their free symbols (and we want to include them)
+        // - If the spec.circuit is a single free symbol, it isn't counter
+        // - the ordering isn't ideal
+        // Due to these issues we recompute the free symbols 'by hand'
+        let mut free_syms: IndexSet<_> = get_free_symbols(&spec.circuit)
+            .iter()
+            .filter(|x| !spec.arg_specs.iter().any(|arg_spec| &arg_spec.symbol == *x))
+            .cloned()
+            .collect();
+        for n in &nodes {
+            free_syms.extend(get_free_symbols(&**n).iter().cloned());
+        }
+        substitute_cached_info.free_symbols = free_syms;
+
+        let substitute_cached_info = Box::new(substitute_cached_info);
 
         let out = Self {
             nodes,
             spec,
-            cached_substituted,
             name,
+            substitute_cached_info,
             info: Default::default(),
         };
         out.initial_init_info()
@@ -762,7 +860,7 @@ impl CircuitNodeHashItems for Module {
 
 impl CircuitNodeSetNonHashInfo for Module {
     fn set_non_hash_info(&mut self) -> Result<()> {
-        *self.info_mut() = self.cached_substituted.info().clone();
+        *self.info_mut() = (*self.substitute_cached_info).clone();
         Ok(())
     }
 }
@@ -998,22 +1096,58 @@ impl Module {
     ) -> CircuitRc {
         self.spec
             .substitute(&self.nodes, name_prefix, name_suffix)
-            .map(|x| x.rename(self.name_cloned()))
             .expect("constructor should ensure this works")
+            .rename(self.name_cloned())
+    }
+}
+
+#[derive(Default)]
+struct AllModuleSubstituter {
+    cache: FastUnboundedCache<HashBytes, CircuitRc>,
+}
+
+impl AllModuleSubstituter {
+    #[apply(cached_method)]
+    #[self_id(self_)]
+    #[key(circuit.info().hash)]
+    #[cache_expr(cache)]
+    pub fn substitute_all_modules(&mut self, circuit: CircuitRc) -> CircuitRc {
+        let old_children: Vec<CircuitRc> = circuit.children().collect();
+        let mut new_children: Vec<_> = old_children
+            .into_iter()
+            .map(|c| self_.substitute_all_modules(c))
+            .collect();
+        if let Some(m) = circuit.as_module() {
+            // it's important that we don't construct a module here - the
+            // module constructor uses the global ALL_MODULE_SUBSTITUTOR so
+            // this would panic on double borrowing the ref cell.
+            let nodes = new_children.split_off(1);
+            let new_spec_circuit = new_children.pop().unwrap();
+            ModuleSpec {
+                circuit: new_spec_circuit,
+                arg_specs: m.spec.arg_specs.clone(),
+            }
+            // we could speed up this substitute invocation probably
+            // currently this is overall O(m*n) where m is number of mods and n is number of circs
+            // but cached to O(n) per each module
+            // and we could instead choose to make this O(n) but with worse caching.
+            // Not clear if this would be slower or faster overall.
+            .substitute(&nodes, None, None)
+            .expect("module constructor should ensure this works")
+            .rename(m.name_cloned())
+        } else {
+            circuit.map_children_unwrap_idxs(|i| new_children[i].clone())
+        }
     }
 }
 
 #[pyfunction]
 pub fn substitute_all_modules(circuit: CircuitRc) -> CircuitRc {
-    deep_map_unwrap(circuit, |c| match &**c {
-        Circuit::Module(mn) => mn.substitute(None, None),
-        _ => c.clone(),
-    })
+    ALL_MODULE_SUBSTITUTOR.with(|sub| sub.borrow_mut().substitute_all_modules(circuit))
 }
 
 #[pyfunction]
 pub fn conform_all_modules(circuit: CircuitRc) -> CircuitRc {
-    // this has to be preorder so that we resolve to outermost sizes
     deep_map_preorder_unwrap(circuit, |c| match &**c {
         Circuit::Module(mn) => mn.conform_to_input_shapes().rc(),
         _ => c.clone(),
@@ -1039,137 +1173,89 @@ pub fn any_children_with_symbolic_sizes(circuit: CircuitRc) -> bool {
     !get_children_with_symbolic_sizes(circuit).is_empty()
 }
 
-// adhoc anyfound
-// TODO: do this batchedly over symbols while still caching individually?
-#[derive(Clone, Debug)]
-pub struct HasFreeSym {
-    cache: cached::UnboundCache<(HashBytes, HashBytes), bool>,
-}
-impl Default for HasFreeSym {
-    fn default() -> Self {
-        Self {
-            cache: cached::UnboundCache::new(),
-        }
-    }
-}
-impl HasFreeSym {
-    #[apply(cached_method)]
-    #[self_id(self_)]
-    #[key((circuit.info().hash, sym.info().hash))]
-    #[cache_expr(cache)]
-    pub fn has_free_sym(&mut self, circuit: &CircuitRc, sym: &Symbol) -> bool {
-        if circuit.as_symbol() == Some(sym) {
-            return true;
-        }
-
-        if let Some(m) = circuit.as_module() {
-            if m.spec
-                .arg_specs
-                .iter()
-                .any(|arg_spec| &arg_spec.symbol == sym)
-            {
-                // skip spec circuit if bound
-                return m.nodes.iter().any(|c| self_.has_free_sym(&c, sym));
-            }
-        }
-
-        circuit.children().any(|c| self_.has_free_sym(&c, sym))
-    }
-}
-
-/// uses global cache
 pub fn has_free_sym(circuit: &CircuitRc, sym: &Symbol) -> bool {
-    HAS_FREE_SYM.with(|has_free_sym| has_free_sym.borrow_mut().has_free_sym(circuit, sym))
-}
-
-// adhoc anyfound
-// TODO: do this batchedly over symbols while still caching individually?
-#[derive(Clone, Debug)]
-pub struct CheckValidSub {
-    cache: cached::UnboundCache<(HashBytes, HashBytes, HashBytes), ()>,
-}
-impl Default for CheckValidSub {
-    fn default() -> Self {
-        Self {
-            cache: cached::UnboundCache::new(),
-        }
+    if let Some(other) = circuit.as_symbol() {
+        other == sym
+    } else {
+        circuit.info().free_symbols.contains(sym)
     }
 }
-impl CheckValidSub {
-    #[apply(cached_method)]
-    #[self_id(self_)]
-    #[key((circuit.info().hash, sym.info().hash, sub_for_sym.info().hash))]
-    #[use_try]
-    #[cache_expr(cache)]
-    pub fn check_valid_sub(
-        &mut self,
-        circuit: &CircuitRc,
-        sym: &Symbol,
-        sub_for_sym: &CircuitRc,
-    ) -> Result<()> {
-        if let Some(sym_here) = circuit.as_symbol() {
-            if sym_here != sym && sym_here.ident() == sym.ident() {
-                bail!(
-                    SubstitutionError::FoundNEQFreeSymbolWithSameIdentification {
-                        sym: sym.clone(),
-                        matching_sym: sym_here.clone()
-                    }
-                )
+
+pub fn intersect_free_syms(circuit: &CircuitRc, syms: &[&Symbol]) -> bool {
+    if let Some(other) = circuit.as_symbol() {
+        syms.contains(&other)
+    } else {
+        syms.iter()
+            .any(|x| circuit.info().free_symbols.contains(*x))
+    }
+}
+
+pub fn recur_check_no_free_bound_by_nested(
+    circ: CircuitRc,
+    syms: &[&Symbol],
+    nodes: &[&CircuitRc],
+    seen: &mut HashSet<HashBytes>,
+) -> Result<()> {
+    if seen.insert(circ.info().hash) {
+        if let Some(m) = circ.as_module() {
+            // first check if we can term early (maybe not actually optimization)
+            if !intersect_free_syms(&circ, syms) {
+                return Ok(());
             }
-        }
 
-        circuit
-            .non_free_children() // do this first so we error at inner most module!
-            .map(|c| self_.check_valid_sub(&c, sym, sub_for_sym))
-            .collect::<Result<()>>()?;
-
-        if let Some(m) = circuit.as_module() {
-            if m.spec
-                .arg_specs
+            // filter out bound symbols for recurring into
+            let (syms, nodes): (Vec<_>, Vec<&CircuitRc>) = syms
                 .iter()
-                .any(|arg_spec| &arg_spec.symbol == sym)
-            {
-                // skip spec circuit if bound
-                return m
-                    .nodes
-                    .iter()
-                    .map(|c| self_.check_valid_sub(&c, sym, sub_for_sym))
-                    .collect();
-            }
-            // otherwise, we might sub into the spec circuit
-            if has_free_sym(&m.spec.circuit, sym) {
-                let used_args = are_args_used(sub_for_sym, &m.spec.arg_specs);
+                .zip(nodes)
+                .filter_map(|(s, n)| {
+                    for arg_spec in &m.spec.arg_specs {
+                        if &arg_spec.symbol == *s {
+                            return None;
+                        }
+                    }
+                    Some((s, n))
+                })
+                .unzip();
 
-                if used_args.iter().any(|x| *x) {
-                    bail!(SubstitutionError::CircuitHasFreeSymsBoundByNestedModule {
-                        circ: sub_for_sym.clone(),
-                        sym: sym.clone(),
-                        nested_module: m.clone(),
-                        bound_by_nested: m
-                            .spec
-                            .arg_specs
-                            .iter()
-                            .zip(used_args)
-                            .filter_map(|(arg_spec, used)| used.then(|| arg_spec.symbol.clone()))
-                            .collect()
-                    })
+            if intersect_free_syms(&m.spec.circuit, &syms) {
+                // fail earlier if possible
+                recur_check_no_free_bound_by_nested(m.spec.circuit.clone(), &syms, &nodes, seen)?;
+
+                // then check valid
+                for (&sub_for_sym, &sym) in nodes.iter().zip(&syms) {
+                    if m.spec
+                        .arg_specs
+                        .iter()
+                        .any(|x| has_free_sym(sub_for_sym, &x.symbol))
+                    {
+                        bail!(SubstitutionError::CircuitHasFreeSymsBoundByNestedModule {
+                            circ: sub_for_sym.clone(),
+                            sym: sym.clone(),
+                            nested_module: m.clone(),
+                            bound_by_nested: m
+                                .spec
+                                .arg_specs
+                                .iter()
+                                .filter_map(|arg_spec| has_free_sym(sub_for_sym, &arg_spec.symbol)
+                                    .then(|| arg_spec.symbol.clone()))
+                                .collect()
+                        })
+                    }
                 }
             }
         }
 
-        Ok(())
+        circ.non_free_children()
+            .map(|child| recur_check_no_free_bound_by_nested(child, syms, nodes, seen))
+            .collect::<Result<Vec<_>>>()?;
     }
+    Ok(())
 }
 
-pub fn check_valid_sub(circuit: &CircuitRc, sym: &Symbol, sub_for_sym: &CircuitRc) -> Result<()> {
-    CHECK_VALID_SUB.with(|check_valid_sub| {
-        check_valid_sub
-            .borrow_mut()
-            .check_valid_sub(circuit, sym, sub_for_sym)
-    })
-}
-
-const SAME_IDENT_P1 :&str = "Substituting for a symbol when there is a not-equal free symbol with the same name and uuid is not allowed (aka, a near miss).";
+const SAME_IDENT_P1: &str = concat!(
+    "Substituting for a symbol when there is a not-equal free symbol",
+    " with the same name and uuid is not allowed (aka, a near miss)."
+);
 const SAME_IDENT_P2: &str = "This is caused by having a different shape (or different named axes).";
 
 #[apply(python_error_exception)]
@@ -1193,28 +1279,31 @@ pub enum SubstitutionError {
 }
 
 use std::cell::RefCell;
-// TODO: maybe cache check_unique_arg_names here?
-// TODO: avoid caching arrays forever!
-// TODO: make modules much faster by avoiding exponential time blow up!
 thread_local! {
     static MODULE_EXPANSIONS: RefCell<
         HashMap<
-            (ModuleSpecHashable, Vec<HashBytes>, Option<String>),
-            CircuitRc, // Weak<Circuit>
+            (ModuleSpecHashable, Vec<HashBytes>, Option<String>, Option<String>),
+            CircuitRc,
         >,
     > = RefCell::new(HashMap::default());
     static MODULE_EXPANSIONS_SHAPE: RefCell<
         HashMap<
             (ModuleSpecHashable, Vec<Shape>),
-            (
-                CircuitRc, // Weak<Circuit>
-                Vec<HashBytes>,
-            ),
+            (CircuitRc, Vec<HashBytes>),
         >,
     > = RefCell::new(HashMap::default());
     static MODULE_EXPANSIONS_AXIS_MAPS: RefCell<
         HashMap<(ModuleSpecHashable, Vec<Shape>), Vec<Vec<Option<usize>>>>,
     > = RefCell::new(HashMap::default());
-    static CHECK_VALID_SUB: RefCell<CheckValidSub> = RefCell::new(Default::default());
-    static HAS_FREE_SYM: RefCell<HasFreeSym> = RefCell::new(Default::default());
+    static NO_FREE_BOUND_BY_NESTED: RefCell<
+        HashSet<(ModuleSpecHashable, Vec<HashBytes>)>,
+    > = RefCell::new(HashSet::default());
+    static ALL_MODULE_SUBSTITUTOR: RefCell<AllModuleSubstituter> = RefCell::new(Default::default());
+}
+
+#[pyfunction]
+pub fn clear_module_circuit_caches() {
+    MODULE_EXPANSIONS.with(|f| f.borrow_mut().clear());
+    MODULE_EXPANSIONS_SHAPE.with(|f| f.borrow_mut().clear());
+    ALL_MODULE_SUBSTITUTOR.with(|f| *f.borrow_mut() = Default::default());
 }
