@@ -1269,14 +1269,11 @@ pub fn get_child_comments(parent: &Circuit, child_pos: usize) -> Vec<String> {
         Circuit::StoredCumulantVar(scv) => {
             format!("K{}", scv.cumulants.keys().nth(child_pos).unwrap())
         }
-        Circuit::Module(module) => {
+        Circuit::Module(_) => {
             if child_pos == 0 {
                 "Spec".to_owned()
-            } else if child_pos % 2 == 1 {
-                "".to_owned()
             } else {
-                let spec_pos = (child_pos - 2) / 2;
-                "Arg ".to_owned() + module.spec.arg_specs[spec_pos].symbol.name().unwrap_or("")
+                "".to_owned()
             }
         }
         _ => "".to_owned(),
