@@ -10,13 +10,13 @@ use rr_util::{fn_struct, py_types::MaybeNotSet, simple_default, util::hash_to_st
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
 use crate::{
-    opaque_iterative_matcher::OpaqueIterativeMatcherVal,
+    opaque_iterative_matcher::{get_opaque_type_matcher, OpaqueIterativeMatcherVal},
     prelude::*,
     print::{
         circuit_to_computability_color_code, circuit_to_size_color_code, get_child_comments,
         op_traversal_per, write_comment, CircuitCommenter, PrintOptions,
     },
-    HashBytes,
+    CircuitType, HashBytes,
 };
 
 #[derive(FromPyObject, Clone)]
@@ -38,7 +38,7 @@ simple_default!(PrintOptionsBase {
     PrintOptionsBase::PrintOptions(Default::default())
 });
 
-fn_struct!(pub CircuitToColor:Fn(circuit:CircuitRc)->String);
+fn_struct!(pub CircuitToColor:Fn(circuit:CircuitRc)->Option<String>);
 
 #[pyclass]
 #[derive(Clone)]
@@ -49,11 +49,13 @@ pub struct PrintHtmlOptions {
     #[pyo3(get, set)]
     pub display_copy_button: bool,
     #[pyo3(get, set)]
-    pub colorer: Option<CircuitToColor>,
+    pub colorer: CircuitToColor,
     #[pyo3(get, set)]
     pub primary_color: String,
     #[pyo3(get, set)]
-    pub display_serial_numbers: bool,
+    pub use_serial_numbers: bool,
+    #[pyo3(get, set)]
+    pub number_leaves: bool,
     #[pyo3(get, set)]
     pub keep_all_cells: bool,
     pub overall_traversal: Option<OpaqueIterativeMatcherVal>,
@@ -69,9 +71,10 @@ impl Default for PrintHtmlOptions {
             shape_only_when_necessary: true,
             traversal: Some(OpaqueIterativeMatcherVal::for_end_depth(3)),
             display_copy_button: true,
-            colorer: Some(PrintHtmlOptions::type_colorer()),
+            colorer: PrintHtmlOptions::type_colorer(),
             primary_color: "lightgray".to_owned(),
-            display_serial_numbers: false,
+            use_serial_numbers: false,
+            number_leaves: false,
             keep_all_cells: true,
             overall_traversal: Some(OpaqueIterativeMatcherVal::never()),
             comment_arg_names: true,
@@ -103,8 +106,12 @@ impl PrintHtmlOptions {
                 NodeJsonData {
                     name: circ.name_cloned().unwrap_or("".to_owned()),
                     desc,
-                    color: match &selfy.colorer {
-                        Some(col_fn) => col_fn.call(circ.clone())?,
+                    color: match selfy
+                        .colorer
+                        .call(circ.clone())
+                        .context("colorer errored in html printer")?
+                    {
+                        Some(col) => col,
                         None => selfy.primary_color.clone(),
                     },
                     children: circ
@@ -115,13 +122,15 @@ impl PrintHtmlOptions {
                 },
             );
 
-            let new_traversal_per =
-                if let Some(new_traversal_per) = op_traversal_per(traversal, circ.crc())? {
-                    new_traversal_per
-                } else {
-                    // all finished case
-                    return Ok(());
-                };
+            let new_traversal_per = if let Some(new_traversal_per) =
+                op_traversal_per(traversal, circ.crc())
+                    .context("traversal errored in html printer")?
+            {
+                new_traversal_per
+            } else {
+                // all finished case
+                return Ok(());
+            };
 
             for (i, (child, new_traversal)) in circ.children().zip(new_traversal_per).enumerate() {
                 let child_specific_commenters = if selfy.comment_arg_names {
@@ -150,6 +159,7 @@ impl PrintHtmlOptions {
             nodes_to_expand: &mut Vec<HashBytes>,
             expanded_nodes: &mut HashSet<HashBytes>,
             data: &mut HashMap<HashBytes, NodeJsonData>,
+            runnning_position_number: &mut usize,
             selfy: &PrintHtmlOptions,
             traversal: Option<OpaqueIterativeMatcherVal>,
         ) -> Result<()> {
@@ -158,10 +168,13 @@ impl PrintHtmlOptions {
             }
             expanded_nodes.insert(circ.info().hash);
 
+            let new_traversal_per_op = op_traversal_per(traversal, circ.crc())?;
+
             let mut desc = String::new();
-            if selfy.display_serial_numbers {
-                desc.push_str(&(expanded_nodes.len() - 1).to_string());
+            if selfy.use_serial_numbers {
+                desc.push_str(&runnning_position_number.to_string());
                 desc.push(' ');
+                *runnning_position_number += 1;
             }
             if let Some(old_data) = data.get_mut(&circ.info().hash) {
                 desc.push_str(&old_data.desc);
@@ -170,13 +183,13 @@ impl PrintHtmlOptions {
                 bail!(anyhow!("overall traversal does not cover traversal."))
             }
 
-            let new_traversal_per =
-                if let Some(new_traversal_per) = op_traversal_per(traversal, circ.crc())? {
-                    new_traversal_per
-                } else {
-                    // all finished case
-                    return Ok(());
-                };
+            let new_traversal_per = if let Some(new_traversal_per) = new_traversal_per_op {
+                new_traversal_per
+            } else {
+                // all finished case
+                return Ok(());
+            };
+            // below is quadratic time due to contains on vec
             if !nodes_to_expand.contains(&circ.info().hash) {
                 nodes_to_expand.push(circ.info().hash);
             }
@@ -186,18 +199,21 @@ impl PrintHtmlOptions {
                     nodes_to_expand,
                     expanded_nodes,
                     data,
+                    runnning_position_number,
                     selfy,
                     new_traversal.clone(),
                 )?;
             }
             Ok(())
         }
+        let mut runnning_position_number = 0;
         for circuit in circuits {
             recurse(
                 circuit,
                 nodes_to_expand,
                 &mut expanded_nodes,
                 data,
+                &mut runnning_position_number,
                 &self,
                 self.traversal.clone(),
             )?;
@@ -221,9 +237,10 @@ impl PrintHtmlOptions {
         shape_only_when_necessary = "PrintHtmlOptions::default().shape_only_when_necessary",
         traversal = "PrintHtmlOptions::default().traversal.unwrap()",
         display_copy_button = "PrintHtmlOptions::default().display_copy_button",
-        colorer = "PrintHtmlOptions::default().colorer.unwrap()",
+        colorer = "PrintHtmlOptions::default().colorer",
         primary_color = "PrintHtmlOptions::default().primary_color",
-        display_serial_numbers = "PrintHtmlOptions::default().display_serial_numbers",
+        use_serial_numbers = "PrintHtmlOptions::default().use_serial_numbers",
+        number_leaves = "PrintHtmlOptions::default().number_leaves",
         keep_all_cells = "PrintHtmlOptions::default().keep_all_cells",
         overall_traversal = "PrintHtmlOptions::default().overall_traversal.unwrap()",
         comment_arg_names = "PrintOptions::default().comment_arg_names",
@@ -233,9 +250,10 @@ impl PrintHtmlOptions {
         shape_only_when_necessary: bool,
         traversal: Option<OpaqueIterativeMatcherVal>,
         display_copy_button: bool,
-        colorer: Option<CircuitToColor>,
+        colorer: CircuitToColor,
         primary_color: String,
-        display_serial_numbers: bool,
+        use_serial_numbers: bool,
+        number_leaves: bool,
         keep_all_cells: bool,
         overall_traversal: Option<OpaqueIterativeMatcherVal>,
         comment_arg_names: bool,
@@ -247,7 +265,8 @@ impl PrintHtmlOptions {
             display_copy_button,
             colorer,
             primary_color,
-            display_serial_numbers,
+            use_serial_numbers,
+            number_leaves,
             keep_all_cells,
             overall_traversal,
             comment_arg_names,
@@ -259,10 +278,11 @@ impl PrintHtmlOptions {
     #[args(
         shape_only_when_necessary = "Default::default()",
         traversal = "Default::default()",
-        colorer = "Default::default()",
+        colorer = "MaybeNotSet(None)",
         primary_color = "Default::default()",
         display_copy_button = "Default::default()",
-        display_serial_numbers = "Default::default()",
+        use_serial_numbers = "Default::default()",
+        number_leaves = "Default::default()",
         keep_all_cells = "Default::default()",
         overall_traversal = "Default::default()",
         comment_arg_names = "Default::default()",
@@ -273,9 +293,10 @@ impl PrintHtmlOptions {
         shape_only_when_necessary: MaybeNotSet<bool>,
         traversal: MaybeNotSet<Option<OpaqueIterativeMatcherVal>>,
         display_copy_button: MaybeNotSet<bool>,
-        colorer: MaybeNotSet<Option<CircuitToColor>>,
+        colorer: MaybeNotSet<CircuitToColor>,
         primary_color: MaybeNotSet<String>,
-        display_serial_numbers: MaybeNotSet<bool>,
+        use_serial_numbers: MaybeNotSet<bool>,
+        number_leaves: MaybeNotSet<bool>,
         keep_all_cells: MaybeNotSet<bool>,
         overall_traversal: MaybeNotSet<Option<OpaqueIterativeMatcherVal>>,
         comment_arg_names: MaybeNotSet<bool>,
@@ -290,9 +311,8 @@ impl PrintHtmlOptions {
             display_copy_button: display_copy_button.0.unwrap_or(cloned.display_copy_button),
             colorer: colorer.0.unwrap_or(cloned.colorer),
             primary_color: primary_color.0.unwrap_or(cloned.primary_color),
-            display_serial_numbers: display_serial_numbers
-                .0
-                .unwrap_or(cloned.display_serial_numbers),
+            use_serial_numbers: use_serial_numbers.0.unwrap_or(cloned.use_serial_numbers),
+            number_leaves: number_leaves.0.unwrap_or(cloned.number_leaves),
             keep_all_cells: keep_all_cells.0.unwrap_or(cloned.keep_all_cells),
             overall_traversal: overall_traversal.0.unwrap_or(cloned.overall_traversal),
             comment_arg_names: comment_arg_names.0.unwrap_or(cloned.comment_arg_names),
@@ -396,6 +416,11 @@ impl PrintHtmlOptions {
         } else {
             "var displayCopyButtontoggle___js___ = false;"
         });
+        html.push_str(if self.number_leaves {
+            "var numberLeaf___js___ = true;"
+        } else {
+            "var numberLeaf___js___ = false;"
+        });
 
         // Jupyter just keeps adding new html elements, so we have to create divs
         // with unique ids. Seed is the suffix of every id created in one block.
@@ -412,7 +437,7 @@ impl PrintHtmlOptions {
         html.push_str(CSS);
 
         html.push_str(&format!(
-            "<div class=\"mainUL\" id=\"{seed}\"><ul></ul></div>",
+            "<div class=\"mainUL\" id=\"s{seed}\"><ul></ul></div>",
         ));
         html.push_str(JS);
 
@@ -448,6 +473,7 @@ impl PrintHtmlOptions {
             MaybeNotSet(None),
             MaybeNotSet(None),
             MaybeNotSet(None),
+            MaybeNotSet(None),
         )
     }
 
@@ -462,56 +488,85 @@ impl PrintHtmlOptions {
         .print(circuits)
     }
 
+    #[staticmethod]
+    pub fn type_nest_default(
+        circuit_type: CircuitType,
+        traversal: Option<OpaqueIterativeMatcherVal>,
+    ) -> Self {
+        Self {
+            number_leaves: true,
+            traversal: Some(get_opaque_type_matcher(circuit_type, traversal)),
+            ..Default::default()
+        }
+    }
+
+    #[staticmethod]
+    pub fn einsum_nest_default(traversal: Option<OpaqueIterativeMatcherVal>) -> Self {
+        Self::type_nest_default(CircuitType::Einsum, traversal)
+    }
+
+    #[staticmethod]
+    pub fn add_nest_default(traversal: Option<OpaqueIterativeMatcherVal>) -> Self {
+        Self::type_nest_default(CircuitType::Add, traversal)
+    }
+
     #[staticmethod]
     pub fn size_colorer() -> CircuitToColor {
         CircuitToColor::new_dyn(Box::new(|c| {
-            Ok(color_from_colorcode(
-                circuit_to_size_color_code(c).unwrap_or(0), // TODO: fix me
-            ))
+            Ok(circuit_to_size_color_code(c).map(|c| color_from_colorcode(c)))
         }))
     }
 
     #[staticmethod]
     pub fn type_colorer() -> CircuitToColor {
         CircuitToColor::new_dyn(Box::new(|circuit: CircuitRc| {
-            let mut hasher = std::collections::hash_map::DefaultHasher::default();
-            circuit.variant_string().hash(&mut hasher);
-            Ok(color_from_hash(hasher.finish() as usize))
+            Ok(Some(match &(**circuit) {
+                Circuit::Einsum(_) => color_from_colorcode(0),
+                Circuit::Add(_) => color_from_colorcode(1),
+                Circuit::Concat(_) => color_from_colorcode(2),
+                Circuit::GeneralFunction(_) => color_from_colorcode(3),
+                Circuit::Array(_) => color_from_colorcode(4),
+                Circuit::DiscreteVar(_) => color_from_colorcode(5),
+                _ => {
+                    let mut hasher = rustc_hash::FxHasher::default();
+                    circuit.variant_string().hash(&mut hasher);
+                    color_from_hash(hasher.finish() as usize)
+                }
+            }))
         }))
     }
 
     #[staticmethod]
     pub fn hash_colorer() -> CircuitToColor {
         CircuitToColor::new_dyn(Box::new(|circuit: CircuitRc| {
-            Ok(color_from_hash(circuit.info().hash_usize()))
+            Ok(Some(color_from_hash(circuit.info().hash_usize())))
         }))
     }
     #[staticmethod]
     pub fn fixed_color(color: String) -> CircuitToColor {
-        CircuitToColor::new_dyn(Box::new(move |_circuit: CircuitRc| Ok(color.clone())))
+        CircuitToColor::new_dyn(Box::new(move |_circuit: CircuitRc| Ok(Some(color.clone()))))
     }
 
     #[staticmethod]
     pub fn computability_colorer() -> CircuitToColor {
         CircuitToColor::new_dyn(Box::new(|c| {
-            Ok(color_from_colorcode(
-                circuit_to_computability_color_code(c).unwrap_or(0), // TODO: fix me
-            ))
+            Ok(circuit_to_computability_color_code(c).map(|c| color_from_colorcode(c)))
         }))
     }
 }
 
 fn color_from_hash(hash: usize) -> String {
-    let hue = (hash % (255 / 10)) * 10;
+    let hue = (hash % (360 / 20)) * 20;
     return format!("hsl({hue}, 90%, 60%)");
 }
 
-const COLORCODE_TO_RGB: [&str; 5] = [
-    "rgb(1,1,1)",
+const COLORCODE_TO_RGB: [&str; 6] = [
     "rgb(222,56,43)",
     "rgb(57,181,74)",
     "rgb(255,199,6)",
     "rgb(0,111,184)",
+    "rgb(118,38,113)",
+    "rgb(44,181,233)",
 ];
 
 fn color_from_colorcode(code: usize) -> String {
@@ -669,7 +724,8 @@ const JS: &str = r#"
 
     // Example of the generated inner HTML:
     // <span class="caret" onclick="toggle('1')"></span>
-    // <span style="color: hsl(70, 100%, 70%)">1 'I k2 logits_diff, true I' [t: Cumulant] [s: ()]</span>
+    // <pre style="color: hsl(70, 100%, 70%)">1 'I k2 logits_diff, true I' [t: Cumulant] [s: ()]</pre>
+    // <pre style="color: magenta"></pre>
     // <button onclick="toClipboard('I k2 logits_diff, true I')">📋</button>
     // <ul class="nested" style="border-left: 3px solid hsl(70, 100%, 70%)"></ul>;
 
@@ -702,7 +758,7 @@ const JS: &str = r#"
       childData.color +
       `">` +
       desc +
-      `</pre>` +
+      `</pre><pre style="color:magenta"></pre>` +
       copyButton +
       `<ul class="nested" style="border-left: 3px solid ` +
       childData.color +
@@ -717,14 +773,6 @@ const JS: &str = r#"
     childData.displayedCount += 1;
     getLastChild(getElement(parent_id)).appendChild(child);
   }
-  rootstoggle___js___.forEach(function (root) {
-    console.log(root);
-    addChildtoggle___js___(seedtoggle___js___, root);
-  });
-  toExpandtoggle___js___.forEach(function (expandElt) {
-    console.log(expandElt);
-    toggletoggle___js___(expandElt + '@0@' + seedtoggle___js___);
-  });
 
   function toClipboard(copyText) {
     navigator.clipboard.writeText(copyText);
@@ -759,5 +807,24 @@ const JS: &str = r#"
       }
     }
   }
+  function updateLeafNumbers___js___() {
+    if (!numberLeaf___js___) return;
+    document.querySelectorAll('#s' + seedtoggle___js___ + ' .nested.active').forEach(function (d, i) {
+      d.parentNode.children[2].innerText = '';
+    });
+    document.querySelectorAll('#s' + seedtoggle___js___ + ' .nested:empty').forEach(function (d, i) {
+      d.parentNode.children[2].innerText = ' # ' + i;
+    });
+  }
+  
+  rootstoggle___js___.forEach(function (root) {
+    console.log(root);
+    addChildtoggle___js___("s"+seedtoggle___js___, root);
+  });
+  toExpandtoggle___js___.forEach(function (expandElt) {
+    console.log(expandElt);
+    toggletoggle___js___(expandElt + '@0@' + seedtoggle___js___);
+  });
+  updateLeafNumbers___js___();
 </script>
 "#;
