@@ -322,7 +322,7 @@ pub fn einsum_nest_path(einsum: &Einsum, path: Vec<Vec<usize>>) -> Einsum {
 pub fn einsum_nest_optimize_py(einsum: &Einsum) -> Option<Einsum> {
     einsum_nest_optimize(
         einsum,
-        &mut OptimizationContext::new_settings_circuit(Default::default(), einsum.crc()),
+        &mut OptimizationContext::new_settings(Default::default()),
     )
 }
 
@@ -907,15 +907,20 @@ pub fn distribute_once_raw(
     einsum: &Einsum,
     operand_idx: usize,
     do_broadcasts: bool,
-    suffix: Option<String>,
+    prefix: Option<&str>,
+    suffix: Option<&str>,
     mut call_on_sub: impl FnMut(usize, Einsum) -> Result<CircuitRc>,
 ) -> Result<Add> {
     let mut on_sub = |child_idx, node: CircuitRc, einsum: Einsum| -> Result<CircuitRc> {
-        let new_name = suffix
-            .as_ref()
-            .map(|s| node.name().map(|n_s| format!("{n_s}{s}")))
-            .flatten();
-        call_on_sub(child_idx, einsum.rename(new_name))
+        let einsum = if prefix.is_some() || suffix.is_some() {
+            let suffix = suffix.unwrap_or("");
+            let prefix = prefix.unwrap_or("");
+
+            einsum.rename(node.name().map(|n_s| format!("{prefix}{n_s}{suffix}")))
+        } else {
+            einsum
+        };
+        call_on_sub(child_idx, einsum)
     };
     match einsum.args[operand_idx].0.as_add() {
         Some(add_node) => {
@@ -955,11 +960,12 @@ pub fn distribute_once_raw(
     }
 }
 
-#[pyfunction]
+#[pyfunction(do_broadcasts = "true", prefix = "None", suffix = "None")]
 pub fn distribute_once(
     einsum: &Einsum,
     operand_idx: usize,
     do_broadcasts: bool,
+    prefix: Option<String>,
     suffix: Option<String>,
 ) -> Result<Add> {
     if operand_idx >= einsum.args.len() {
@@ -972,13 +978,17 @@ pub fn distribute_once(
         einsum,
         operand_idx,
         do_broadcasts,
-        suffix,
+        prefix.as_deref(),
+        suffix.as_deref(),
         |_, c| Ok(c.rc()),
     )
 }
 
 pub fn distribute_once_op(einsum: &Einsum, operand_idx: usize, do_broadcasts: bool) -> Option<Add> {
-    distribute_once_raw(einsum, operand_idx, do_broadcasts, None, |_, c| Ok(c.rc())).ok()
+    distribute_once_raw(einsum, operand_idx, do_broadcasts, None, None, |_, c| {
+        Ok(c.rc())
+    })
+    .ok()
 }
 
 /// distribute all Adds that are direct children of this einsum
