@@ -186,10 +186,7 @@ pub enum BatchError {
 #[pyfunction]
 #[pyo3(name = "batch_einsum")]
 pub fn batch_einsum_py(einsum: Einsum, settings: OptimizationSettings) -> Result<CircuitRc> {
-    batch_einsum(
-        &einsum,
-        &mut OptimizationContext::new_settings_circuit(settings, einsum.crc()),
-    )
+    batch_einsum(&einsum, &mut OptimizationContext::new_settings(settings))
 }
 
 pub fn batch_einsum(einsum: &Einsum, context: &mut OptimizationContext) -> Result<CircuitRc> {
