@@ -2,7 +2,7 @@ use std::io::prelude::*;
 
 use anyhow::{bail, Context, Result};
 use circuit_base::{
-    flat_concat, get_compatible_dtype,
+    flat_concat,
     module::{any_children_with_symbolic_sizes, conform_all_modules},
     print::{print_circuit_stats, PrintOptions},
     CircuitNode, CircuitRc, ModuleSpec,
@@ -198,15 +198,6 @@ impl OptimizationContext {
             settings,
         }
     }
-    #[staticmethod]
-    pub fn new_settings_circuit(settings: OptimizationSettings, circuit: CircuitRc) -> Self {
-        let mut result = Self {
-            cache: Default::default(),
-            settings,
-        };
-        let dtype = get_compatible_dtype(&circuit);
-        result
-    }
 
     pub fn stringify_logs(&self) -> String {
         format!(
@@ -225,7 +216,7 @@ pub fn optimize_circuit_py(
 ) -> Result<CircuitRc> {
     optimize_circuit(
         circuit.clone(),
-        &mut OptimizationContext::new_settings_circuit(settings, circuit),
+        &mut OptimizationContext::new_settings(settings),
     )
 }
 
@@ -329,7 +320,7 @@ pub fn optimize_to_schedule(
     circuit: CircuitRc,
     settings: OptimizationSettings,
 ) -> Result<Schedule> {
-    let context = &mut OptimizationContext::new_settings_circuit(settings, circuit.clone());
+    let context = &mut OptimizationContext::new_settings(settings);
     let verbose = settings.verbose;
     if verbose > 0 {
         python_println!("Optimizing verbose {}", verbose)
