@@ -80,12 +80,9 @@ pub fn opt_eval_each_subcircuit_until_fail(circuit: CircuitRc, settings: Optimiz
                 if close_result.is_err() {
                     println!("{:?}", close_result);
                     circ.printu();
-                    optimize_circuit(
-                        circ,
-                        &mut OptimizationContext::new_settings_circuit(settings, circuit.clone()),
-                    )
-                    .unwrap()
-                    .printu();
+                    optimize_circuit(circ, &mut OptimizationContext::new_settings(settings))
+                        .unwrap()
+                        .printu();
                 }
             }
         }
