@@ -613,29 +613,30 @@ pub fn deep_heuristic_nest_adds(circ: CircuitRc) -> CircuitRc {
     });
     // let mut stats = (0, 0, 0);
     // let mut add_width_tracker: Vec<usize> = vec![0; 300];
-    let mut intersections: HashMap<Add, CircBloomFilter> = HashMap::default();
-    for circ in &seen_adds {
-        for circ2 in &seen_adds {
-            if circ.1.intersection_count(&circ2.1) >= 2 && circ.0 != circ2.0 {
-                let intersection = Add::new(
-                    circ.0
-                        .nodes
-                        .iter()
-                        .filter(|x| circ2.0.nodes.contains(x))
-                        .cloned()
-                        .collect(),
-                    None,
-                );
-                if intersection.nodes.len() >= 2
-                    && &intersection != &circ.0
-                    && &intersection != &circ2.0
-                {
-                    intersections.insert(intersection.clone(), child_bloom_filter(&intersection));
-                }
-            }
-        }
-    }
-    seen_adds.extend(intersections);
+    // let mut intersections: HashMap<Add, CircBloomFilter> = HashMap::default();
+    // skipping intersections bc we're too slow as it is
+    // for circ in &seen_adds {
+    //     for circ2 in &seen_adds {
+    //         if circ.1.intersection_count(&circ2.1) >= 2 && circ.0 != circ2.0 {
+    //             let intersection = Add::new(
+    //                 circ.0
+    //                     .nodes
+    //                     .iter()
+    //                     .filter(|x| circ2.0.nodes.contains(x))
+    //                     .cloned()
+    //                     .collect(),
+    //                 None,
+    //             );
+    //             if intersection.nodes.len() >= 2
+    //                 && &intersection != &circ.0
+    //                 && &intersection != &circ2.0
+    //             {
+    //                 intersections.insert(intersection.clone(), child_bloom_filter(&intersection));
+    //             }
+    //         }
+    //     }
+    // }
+    // seen_adds.extend(intersections);
     let mut seen_adds_index: HashMap<HashBytes, usize> = seen_adds
         .iter()
         .enumerate()
@@ -645,7 +646,9 @@ pub fn deep_heuristic_nest_adds(circ: CircuitRc) -> CircuitRc {
     //     add_width_tracker[ad.0.nodes.len()] += 1;
     // }
     let mut mapping: HashMap<Add, Add> = HashMap::default();
-    for _ in 0..6 {
+    for _ in 0..2 {
+        // should allow this forever but limiting to 2 bc slow
+        let timeout_timer = std::time::Instant::now();
         let mut to_do: Vec<((Add, CircBloomFilter), (Add, CircBloomFilter))> = vec![];
         for cand_sup in &seen_adds {
             if cand_sup.0.nodes.len() >= 3 {
@@ -664,24 +667,30 @@ pub fn deep_heuristic_nest_adds(circ: CircuitRc) -> CircuitRc {
                         }
                     }
                 }
+                if timeout_timer.elapsed().as_millis() > 1000 {
+                    println!("nest adds timed out");
+                    break;
+                }
             }
         }
         // println!("stats {:?} {:?}", stats, add_width_tracker);
         if to_do.is_empty() {
             break;
         }
+
         for (sup, new_sup) in to_do {
-            let mut seen_index = seen_adds_index.remove(&sup.0.info().hash);
-            let to_insert = (new_sup.0.clone(), new_sup.1.clone());
-            if let Some(si) = seen_index {
-                seen_adds[si] = to_insert;
-            } else {
-                seen_index = Some(seen_adds.len());
-                // add_width_tracker[to_insert.0.nodes.len()] += 1;
-                seen_adds.push(to_insert);
-            }
-            if !seen_adds_index.contains_key(&new_sup.0.info().hash) {
-                seen_adds_index.insert(new_sup.0.info().hash, seen_index.unwrap());
+            let old_sup_index = seen_adds_index
+                .insert(sup.0.info().hash, usize::MAX)
+                .unwrap();
+            if old_sup_index != usize::MAX {
+                if !seen_adds_index.contains_key(&new_sup.0.info().hash) {
+                    seen_adds_index.insert(new_sup.0.info().hash, old_sup_index);
+                    seen_adds[old_sup_index] = new_sup.clone();
+                } else if let Some(popped) = seen_adds.pop() {
+                    // if the new superset is already seen, pop last thing ad put it at old sup index and fix seen_adds_index
+                    seen_adds_index.insert(popped.0.info().hash, old_sup_index);
+                    seen_adds[old_sup_index] = popped;
+                }
             }
             mapping.insert(sup.0.clone(), new_sup.0.clone());
         }
