@@ -24,25 +24,40 @@ pub fn elim_no_input_module(module: &Module) -> Option<CircuitRc> {
     module.nodes.is_empty().then(|| module.spec.circuit.clone())
 }
 
-#[pyfunction(use_elim_no_input_module = "true")]
+#[pyfunction(
+    add_suffix_on_remove_unused = "true",
+    use_elim_no_input_module = "true"
+)]
 #[pyo3(name = "deep_module_remove_unused_inputs")]
 pub fn py_deep_module_remove_unused_inputs(
     circuit: CircuitRc,
+    add_suffix_on_remove_unused: bool,
     use_elim_no_input_module: bool,
 ) -> CircuitRc {
-    deep_module_remove_unused_inputs(circuit, &mut Default::default(), use_elim_no_input_module)
+    deep_module_remove_unused_inputs(
+        circuit,
+        &mut Default::default(),
+        add_suffix_on_remove_unused,
+        use_elim_no_input_module,
+    )
 }
 
 pub fn deep_module_remove_unused_inputs(
     circuit: CircuitRc,
     cache: &mut HashMap<HashBytes, Option<CircuitRc>>,
+    add_suffix_on_remove_unused: bool,
     use_elim_no_input_module: bool,
 ) -> CircuitRc {
     deep_map_op_context(
         circuit.clone(),
         &|c, _| {
             c.as_module().and_then(|m| {
-                module_remove_unused_inputs(m, use_elim_no_input_module).ok()
+                module_remove_unused_inputs(
+                    m,
+                    add_suffix_on_remove_unused,
+                    use_elim_no_input_module,
+                )
+                .ok()
                 // ok case is rank overflow, rare.
             })
         },
@@ -54,6 +69,7 @@ pub fn deep_module_remove_unused_inputs(
 
 pub fn module_remove_unused_inputs_op(
     m: &Module,
+    add_suffix_on_remove_unused: bool,
     use_elim_no_input_module: bool,
 ) -> Result<Option<CircuitRc>> {
     let used_inputs = m.spec.are_args_used();
@@ -78,7 +94,13 @@ pub fn module_remove_unused_inputs_op(
             circuit: m.spec.circuit.clone(),
             arg_specs,
         },
-        m.name().map(|s| format!("{} rem_unused", s)),
+        m.name().map(|s| {
+            if add_suffix_on_remove_unused {
+                format!("{} rem_unused", s)
+            } else {
+                s.to_owned()
+            }
+        }),
     );
 
     let out_batch_shape = m_out.aligned_batch_shape();
@@ -107,12 +129,19 @@ pub fn module_remove_unused_inputs_op(
     Ok(Some(out))
 }
 
-#[pyfunction(use_elim_no_input_module = "true")]
+#[pyfunction(
+    add_suffix_on_remove_unused = "true",
+    use_elim_no_input_module = "true"
+)]
 pub fn module_remove_unused_inputs(
     m: &Module,
+    add_suffix_on_remove_unused: bool,
     use_elim_no_input_module: bool,
 ) -> Result<CircuitRc> {
-    Ok(module_remove_unused_inputs_op(m, use_elim_no_input_module)?.unwrap_or_else(|| m.crc()))
+    Ok(
+        module_remove_unused_inputs_op(m, add_suffix_on_remove_unused, use_elim_no_input_module)?
+            .unwrap_or_else(|| m.crc()),
+    )
 }
 
 #[pyfunction(
