@@ -895,7 +895,7 @@ pub fn scheduled_evaluate(circuit: CircuitRc, settings: OptimizationSettings) ->
     } else {
         circuit_to_schedule(
             circuit.clone(),
-            &mut OptimizationContext::new_settings_circuit(settings, circuit),
+            &mut OptimizationContext::new_settings(settings),
         )
     }?;
     Ok(evaluate_schedule(&schedule)[&schedule.output_circuit.unwrap().0].clone())
