@@ -10,10 +10,7 @@ use rr_util::{
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use smallvec::SmallVec as Sv;
 
-use crate::{
-    scheduled_execution::{PySchedulingOOMManyError, SchedulingOOMError},
-    scheduling_z3::schedule_dag_strategy_ints,
-};
+use crate::{scheduled_execution::SchedulingOOMError, scheduling_z3::schedule_dag_strategy_ints};
 pub struct DagSimpSettings {
     pub same_ratio: f64,
     pub tiny_size: usize,
