@@ -1,7 +1,7 @@
 use std::{
     collections::{BTreeMap, BTreeSet},
     fmt::{self, Debug},
-    ops,
+    iter, ops,
     sync::Arc,
     vec,
 };
@@ -22,7 +22,9 @@ use pyo3::{
 };
 use rr_util::{
     eq_by_big_hash::EqByBigHash,
-    impl_both_by_big_hash, python_error_exception, setup_callable, simple_default, simple_from,
+    impl_both_by_big_hash,
+    py_types::PyCallable,
+    python_error_exception, setup_callable, simple_default, simple_from,
     tensor_util::Slice,
     util::{
         arc_ref_clone, arc_unwrap_or_clone, as_sorted, flip_op_result, split_first_take, transpose,
@@ -44,7 +46,7 @@ pub enum IterativeMatcherFromPy {
     // Finished(Finished),
     IterativeMatcher(IterativeMatcher),
     #[pyo3(transparent)]
-    PyFunc(PyObject),
+    PyMatcherFunc(PyCallable), // goes to matcher!
 }
 
 #[derive(Clone, Debug)]
@@ -58,7 +60,7 @@ pub enum IterativeMatcherData {
     Chains(BTreeSet<ChainItem>),
     All(Vec<IterativeMatcherRc>),
     Raw(RawIterativeMatcher),
-    PyFunc(PyObject),
+    PyFunc(PyCallable), // NOTE: not PyMatcherFunc
 }
 
 setup_callable!(IterativeMatcher, IterativeMatcherData, IterativeMatcherFromPy, match_iterate(circuit : CircuitRc) -> IterateMatchResults);
@@ -93,7 +95,7 @@ impl From<IterativeMatcherFromPy> for IterativeMatcher {
             // we intentionally do a matcher here as the default - if users want
             // a IterativeMatcher pyfunc, they can explicitly use IterativeMatcher
             // factory.
-            IterativeMatcherFromPy::PyFunc(x) => MatcherData::PyFunc(x).into(),
+            IterativeMatcherFromPy::PyMatcherFunc(x) => MatcherData::PyFunc(x).into(),
         }
     }
 }
@@ -306,26 +308,13 @@ impl FilterIterativeMatcher {
             });
         }
 
-        let before_start = self.start_depth.map(|x| self.depth < x).unwrap_or(false);
-        if before_start {
-            return Ok(IterateMatchResults {
-                updated: Some(
-                    Some(
-                        IterativeMatcherData::Filter(FilterIterativeMatcher {
-                            depth: self.depth + 1,
-                            ..self.clone()
-                        })
-                        .into(),
-                    )
-                    .into(),
-                ),
-                found: false,
-            });
-        }
-
         let IterateMatchResults { updated, found } =
             self.iterative_matcher.match_iterate(circuit.clone())?;
 
+        let before_start = self.start_depth.map(|x| self.depth < x).unwrap_or(false);
+
+        let found = found && !before_start;
+
         let reached_end = with_dist_of_end(1);
         if all_finished(&updated)
             || (found && self.term_if_matches)
@@ -338,13 +327,15 @@ impl FilterIterativeMatcher {
             });
         }
 
-        let new_depth = if self.end_depth.is_some() {
+        let needs_depth_update = self.end_depth.is_some() || before_start;
+
+        let new_depth = if needs_depth_update {
             self.depth + 1
         } else {
             self.depth
         };
 
-        let updated = if self.end_depth.is_some() {
+        let updated = if needs_depth_update {
             // we can't keep same if we need to update end_depth
             Some(updated.unwrap_or(Some(self.iterative_matcher.clone()).into()))
         } else {
@@ -967,20 +958,40 @@ impl IterativeMatcher {
     }
 
     #[staticmethod]
-    #[args(first, rest = "*")]
-    pub fn new_chain(first: IterativeMatcherRc, rest: Vec<IterativeMatcherRc>) -> Self {
-        first.chain(rest)
+    #[args(first, rest = "*", must_be_sub = "false")]
+    pub fn new_chain(
+        first: IterativeMatcherRc,
+        rest: Vec<IterativeMatcherRc>,
+        must_be_sub: bool,
+    ) -> Self {
+        first.chain(rest, must_be_sub)
     }
 
     #[staticmethod]
-    #[args(chains = "*")]
-    pub fn new_chain_many(chains: Vec<Vec<IterativeMatcherRc>>) -> Result<Self> {
+    #[args(chains = "*", must_be_sub = "false")]
+    pub fn new_chain_many(chains: Vec<Vec<IterativeMatcherRc>>, must_be_sub: bool) -> Result<Self> {
         Ok(IterativeMatcherData::Chains(
             chains
                 .into_iter()
                 .map(|mut chain| match split_first_take(&mut chain) {
                     None => bail!("Received empty tuple for a chain, we expect len >= 1",),
-                    Some((first, rest)) => Ok(ChainItem::new(first, rest.collect())),
+                    Some((first, rest)) => Ok(ChainItem::new(
+                        first,
+                        rest.map(|x| {
+                            if must_be_sub {
+                                x.filter(
+                                    false,
+                                    Some(1),
+                                    None,
+                                    MatcherFromPyBase::Always(false).into(),
+                                )
+                                .rc()
+                            } else {
+                                x
+                            }
+                        })
+                        .collect(),
+                    )),
                 })
                 .collect::<Result<_>>()?,
         )
@@ -1002,7 +1013,7 @@ impl IterativeMatcher {
 
     #[staticmethod]
     #[pyo3(name = "new_func")]
-    pub(super) fn new_func_py(f: PyObject) -> Self {
+    pub(super) fn new_func_py(f: PyCallable) -> Self {
         IterativeMatcherData::PyFunc(f).into()
     }
 
@@ -1050,28 +1061,25 @@ macro_rules! dup_functions {
 
         }
     } => {
-        // paste is needed due to how pymethods proc macro works
-        paste::paste!{
-            #[pymethods]
-            impl IterativeMatcher {
-                $(
-                    $(#[$($meta_tt)*])*
-                    $vis fn $name(&self, $($arg_name : $arg_ty,)*) $(-> $ret_ty)* {
-                        let $self_ident = self;
-                        $($tt)*
-                    }
-                )*
-            }
+        #[pymethods]
+        impl IterativeMatcher {
+            $(
+                $(#[$($meta_tt)*])*
+                $vis fn $name(&self, $($arg_name : $arg_ty,)*) $(-> $ret_ty)* {
+                    let $self_ident = self;
+                    $($tt)*
+                }
+            )*
+        }
 
-            #[pymethods]
-            impl Matcher {
-                $(
-                    $(#[$($meta_tt)*])*
-                    $vis fn $name(&self, $($arg_name : $arg_ty,)*) $(-> $ret_ty)* {
-                        self.to_iterative_matcher().$name($($arg_name,)*)
-                    }
-                )*
-            }
+        #[pymethods]
+        impl Matcher {
+            $(
+                $(#[$($meta_tt)*])*
+                $vis fn $name(&self, $($arg_name : $arg_ty,)*) $(-> $ret_ty)* {
+                    self.to_iterative_matcher().$name($($arg_name,)*)
+                }
+            )*
         }
 
     };
@@ -1123,26 +1131,30 @@ impl IterativeMatcher {
         ))
     }
 
-    #[args(rest = "*")]
-    pub fn chain(&self, rest: Vec<IterativeMatcherRc>) -> IterativeMatcher {
+    #[args(rest = "*", must_be_sub = "false")]
+    pub fn chain(&self, rest: Vec<IterativeMatcherRc>, must_be_sub: bool) -> IterativeMatcher {
         // TODO: flatten
-        IterativeMatcherData::Chains(
-            [ChainItem::new(self_.clone().into(), rest)]
-                .into_iter()
-                .collect::<BTreeSet<_>>(),
+        Self::new_chain_many(
+            vec![iter::once(self_.clone().rc()).chain(rest).collect()],
+            must_be_sub,
         )
-        .into()
+        .unwrap()
     }
 
-    #[args(rest = "*")]
-    pub fn chain_many(&self, rest: Vec<Vec<IterativeMatcherRc>>) -> IterativeMatcher {
+    #[args(rest = "*", must_be_sub = "false")]
+    pub fn chain_many(
+        &self,
+        rest: Vec<Vec<IterativeMatcherRc>>,
+        must_be_sub: bool,
+    ) -> IterativeMatcher {
         // TODO: flatten
-        IterativeMatcherData::Chains(
+        Self::new_chain_many(
             rest.into_iter()
-                .map(|x| ChainItem::new(self_.clone().into(), x))
+                .map(|x| iter::once(self_.clone().rc()).chain(x).collect())
                 .collect(),
+            must_be_sub,
         )
-        .into()
+        .unwrap()
     }
 
     pub fn children_matcher(&self, child_numbers: BTreeSet<usize>) -> IterativeMatcher {
