@@ -369,7 +369,7 @@ pub fn append_matchers_to_names(
         matcher: Option<IterativeMatcherRc>,
         discard_old_name: bool,
     ) -> Result<CircuitRc> {
-        if matcher.is_none() {
+        if matcher.is_none() || circuit.is_symbol() {
             return Ok(circuit);
         }
         let matcher = matcher.unwrap();
