@@ -18,9 +18,10 @@ pub fn traverse_until_depth(depth: Option<u32>) -> IterativeMatcherRc {
         .rc()
 }
 
-// TODO: improve me: fixed operand index isn't very flexible (though I'm not sure that deep distribute rewrite is that important...)
+// TODO: improve me: fixed operand index isn't very flexible in deep case (though I'm not sure that deep distribute rewrite is that important...)
 #[pyfunction(
     traversal = "traverse_until_depth(Some(1))",
+    prefix = "None",
     suffix = "None",
     allow_partial_distribute = "true",
     do_broadcasts = "true"
@@ -29,6 +30,7 @@ pub fn distribute(
     einsum: Einsum,
     operand_idx: usize,
     traversal: Option<IterativeMatcherRc>,
+    prefix: Option<String>,
     suffix: Option<String>,
     allow_partial_distribute: bool, // noop if can't distribute
     do_broadcasts: bool,
@@ -63,7 +65,8 @@ pub fn distribute(
             &einsum,
             operand_idx,
             do_broadcasts,
-            suffix.clone(),
+            prefix.as_deref(),
+            suffix.as_deref(),
             |i, new_einsum| distribute_rec(new_einsum, traversal_per_child[i].clone()),
         );
         if distributed.is_err() && allow_partial_distribute {
