@@ -7,7 +7,7 @@ use std::{
 
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
-    expand_node::{expand_node, map_replace_expand, ReplaceMapRc},
+    expand_node::{expand_node, MapReplaceExpander, ReplaceMapRc},
     module::are_args_used,
     CircuitNode, CircuitRc, Module, ModuleArgSpec, ModuleSpec, Rearrange, Symbol,
 };
@@ -25,6 +25,7 @@ use pyo3::{
 };
 use rr_util::{
     cached_method,
+    caching::FastUnboundedCache,
     eq_by_big_hash::EqByBigHash,
     fn_struct, impl_eq_by_big_hash, pycall, python_error_exception,
     rearrange_spec::{OpSize, RearrangeSpec, RearrangeSpecError},
@@ -445,7 +446,7 @@ pub struct ModulePusher {
     #[pyo3(get)]
     namer: NestedModuleNamer,
     any_found: AnyFound,
-    cache: cached::UnboundCache<
+    cache: FastUnboundedCache<
         (
             HashBytes,
             IterativeMatcherRc,
@@ -456,6 +457,7 @@ pub struct ModulePusher {
         ),
         (CircuitRc, HashSet<CircuitRc>),
     >,
+    replace_expander: MapReplaceExpander,
 }
 
 impl Default for ModulePusher {
@@ -468,7 +470,8 @@ impl Default for ModulePusher {
             bind_encountered_symbols: true,
             namer: Default::default(),
             any_found: Default::default(),
-            cache: cached::UnboundCache::new(),
+            cache: Default::default(),
+            replace_expander: MapReplaceExpander::new_noop(),
         }
     }
 }
@@ -765,7 +768,9 @@ impl ModulePusher {
         items: &NestedModuleItems,
         extra_replacements: &ReplaceMapRc,
     ) -> Result<CircuitRc> {
-        let finished_circuit = map_replace_expand(circuit.clone(), extra_replacements)?;
+        let finished_circuit = self
+            .replace_expander
+            .replace_expand_with_map(circuit.clone(), extra_replacements)?;
         let rep_expand_extra_dims = finished_circuit.ndim() - circuit.ndim();
 
         let mut updated_extra_replacements = (**extra_replacements).clone();
@@ -808,7 +813,7 @@ impl ModulePusher {
             let (finished_circuit, new_flat) = if rep_expand_extra_dims > 0 {
                 // if we have extra dims, we have to pull the symbols and resub to sync up batching
                 let new_flat = update_flat(items.flat.clone(), rep_expand_extra_dims);
-                let new_finished_circuit = map_replace_expand(
+                let new_finished_circuit = self.replace_expander.replace_expand_with_map(
                     circuit.clone(),
                     &ReplaceMapRc::new(updated_extra_replacements),
                 )?;
@@ -845,7 +850,7 @@ impl ModulePusher {
                         ))
                     })
                     .collect();
-                let new_finished_circuit = map_replace_expand(
+                let new_finished_circuit = self.replace_expander.replace_expand_with_map(
                     circuit.clone(),
                     &ReplaceMapRc::new(updated_extra_replacements),
                 )?;
@@ -976,7 +981,7 @@ impl EqByBigHash for SymbolSetRc {
 impl_eq_by_big_hash!(SymbolSetRc);
 
 struct SymbolExtractor {
-    cache: cached::UnboundCache<
+    cache: FastUnboundedCache<
         (HashBytes, HashBytes, IterativeMatcherRc),
         (CircuitRc, BTreeMap<Symbol, (CircuitRc, ModuleArgSpec)>),
     >,
@@ -986,7 +991,7 @@ struct SymbolExtractor {
 impl Default for SymbolExtractor {
     fn default() -> Self {
         Self {
-            cache: cached::UnboundCache::new(),
+            cache: FastUnboundedCache::default(),
             conform_batch_if_needed: false,
         }
     }
