@@ -1193,15 +1193,16 @@ pub fn nest_flat_einsum_strict_rec(
                         if spec.info.shrink_out_axes {
                             // keep ordering for user simplicity
                             sub_out_axes
-                                .into_iter()
-                                .filter(|x| minimal_out_num_set.contains(x))
+                                .iter()
+                                .filter(|x| minimal_out_num_set.contains(*x))
+                                .cloned()
                                 .collect()
                         } else {
                             sub_out_axes
                         }
                     } else {
                         let mut sub_out_axes: EinsumAxes = out_ints.collect();
-                        sub_out_axes.sort();
+                        sub_out_axes.as_mut_slice().sort();
                         sub_out_axes
                     };
 
@@ -1213,7 +1214,7 @@ pub fn nest_flat_einsum_strict_rec(
                                 expected_len: sub_out_axes.len()
                             })
                         }
-                        sub_out_axes = perm.into_iter().map(|i| sub_out_axes[i as usize]).collect();
+                        sub_out_axes = perm.iter().map(|i| sub_out_axes[*i as usize]).collect();
                     }
 
                     (
