@@ -10,6 +10,7 @@ pub mod eq_by_big_hash;
 pub mod repr;
 #[macro_use]
 pub mod errors_util;
+pub mod char_tokenizer;
 pub mod compact_data;
 pub mod lazy;
 pub mod lru_cache;
@@ -26,6 +27,9 @@ pub mod tensor_util;
 pub mod union_find;
 pub mod util;
 
+pub type IndexSet<T> =
+    indexmap::set::IndexSet<T, core::hash::BuildHasherDefault<rustc_hash::FxHasher>>;
+
 #[macro_export]
 macro_rules! sv {
     [$($tt:tt)*] => {
