@@ -993,6 +993,10 @@ impl TorchDeviceDtypeOp {
         )
     }
 
+    pub fn get_torch_dtype(&self) -> Option<PyObject> {
+        self.dtype.map(TorchDtype::get_torch)
+    }
+
     fn __richcmp__(&self, object: &Self, comp_op: CompareOp) -> bool {
         use_rust_comp(&self, &object, comp_op)
     }
@@ -1072,6 +1076,12 @@ string_enum! {
     )
 }
 
+impl TorchDtype {
+    pub fn get_torch(self) -> PyObject {
+        PY_UTILS.dtype_convert[&self].clone()
+    }
+}
+
 string_enum! {
     TorchDevice,
     None,
@@ -1150,6 +1160,11 @@ impl TorchDeviceDtype {
             String::from(&self.dtype),
         )
     }
+
+    pub fn get_torch_dtype(&self) -> PyObject {
+        self.dtype.get_torch()
+    }
+
     fn __richcmp__(&self, object: &Self, comp_op: CompareOp) -> bool {
         use_rust_comp(&self, &object, comp_op)
     }
