@@ -1,6 +1,6 @@
 use std::{
     collections::BTreeMap,
-    fmt::Debug,
+    fmt::{Debug, Display},
     hash::{BuildHasher, Hash},
     mem,
     rc::Rc,
@@ -282,7 +282,7 @@ pub fn mapping_until_end<T: Eq + Hash + Clone>(x: &T, mapping: &HashMap<T, T>) -
     panic!("mapping_until_end didnt finish");
 }
 
-pub fn apply_fn_until_same<T: Eq + Hash + Clone + Debug, F>(x: &T, f: F) -> T
+pub fn apply_fn_until_same<T: Eq + Hash + Clone + Debug + Display, F>(x: &T, f: F) -> T
 where
     F: FnMut(&T) -> T,
 {
@@ -614,7 +614,7 @@ macro_rules! fn_struct{
         paste::paste!{
             #[derive(Clone,Debug)]
             $vis enum $name{
-                Py(pyo3::PyObject),
+                Py($crate::py_types::PyCallable),
                 Dyn([< $name DynStruct >]),
             }
             impl $name{
@@ -642,13 +642,13 @@ macro_rules! fn_struct{
             }
             impl<'source> pyo3::FromPyObject<'source> for $name{
                 fn extract(inp: &'source pyo3::PyAny) -> pyo3::PyResult<Self> {
-                    Ok(Self::Py(inp.into()))
+                    Ok(Self::Py(inp.extract()?))
                 }
             }
             impl IntoPy<PyObject> for $name{
                 fn into_py(self, py: Python<'_>) -> PyObject {
                     match self{
-                        $name::Py(x)=>x,
+                        $name::Py(x)=>x.into_py(py),
                         $name::Dyn(dynf)=>dynf.into_py(py)
                     }
                 }
