@@ -68,6 +68,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     if !cfg!(target_endian = "little") {
         return PyResult::Err(PyValueError::new_err("tried to build non little endian, rr_util::compact_data::TinyVecU8 relies on little endian"));
     }
+    m.add_class::<rr_util::char_tokenizer::CharTokenizer>()?;
 
     m.add_class::<circuit_base::PyCircuitBase>()?;
 
@@ -85,7 +86,6 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         "GeneralFunctionSpecBase",
         &*circuit_base::generalfunction::PY_WRAP_BASE,
     )?;
-    m.add_class::<rr_util::opt_einsum::EinsumSpec>()?;
     m.add_class::<circuit_base::Einsum>()?;
     m.add_class::<circuit_base::Array>()?;
     m.add_class::<circuit_base::Symbol>()?;
@@ -114,7 +114,10 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<circuit_base::module::Module>()?;
     m.add_class::<circuit_base::module::ModuleSpec>()?;
     m.add_class::<circuit_base::module::ModuleArgSpec>()?;
-    m.add_function(wrap_pyfunction!(circuit_base::module::all_free, m)?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_base::module::py_get_free_symbols,
+        m
+    )?)?;
     m.add_class::<circuit_base::SetSymbolicShape>()?;
     m.add_class::<circuit_base::Tag>()?;
     m.add_class::<circuit_base::DiscreteVar>()?;
@@ -274,7 +277,6 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         circuit_rewrites::circuit_optimizer::optimize_circuit_py,
         m
     )?)?;
-    m.add_function(wrap_pyfunction!(optimize_einsum_spec_cached, m)?)?;
     m.add_function(wrap_pyfunction!(scatter_to_concat, m)?)?;
     m.add_function(wrap_pyfunction!(scheduled_evaluate, m)?)?;
     m.add_function(wrap_pyfunction!(optimize_and_evaluate, m)?)?;
@@ -374,6 +376,10 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         circuit_base::module::conform_all_modules,
         m
     )?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_base::module::clear_module_circuit_caches,
+        m
+    )?)?;
     m.add_function(wrap_pyfunction!(
         circuit_base::module::get_children_with_symbolic_sizes,
         m
